# script d'installation des applications via winget
# executé avant wpkg, il est donc prioritaire sur celui-ci. Idéalement wpkg n'a plus rien à faire ensuite.
#
# on envoie au se4fs un json contenant la liste des applications. Celle-ci est analysée et
#  - on récupère un json contenant la listes des applications à désinstaller
#  - on récupère la liste des applications installer mettre à jour.
# on envoie à nouveau le résultat pour la mise à jour des rapports
#
# si une application définit explicitement une version dans wpkg, on ne la met pas à jour.
# dans le cas contraire on met à jour sysématiquement.


#Check Winget Install
Write-Host "Checking if Winget is installed" -ForegroundColor Yellow
$TestWinget = Get-AppxProvisionedPackage -Online | Where-Object {$_.DisplayName -eq "Microsoft.DesktopAppInstaller"}
If ([Version]$TestWinGet. Version -gt "2022.506.16.0") {
    Write-Host "WinGet is Installed" -ForegroundColor Green
} Else {
    #WebClient
    $dc = New-Object net.webclient
    $dc.UseDefaultCredentials = $true
    $dc.Headers.Add("user-agent", "Internet Explorer")
    $dc.Headers.Add("X-FORMS_BASED_AUTH_ACCEPTED", "f")

    #temp folder
    $InstallerFolder = $(Join-Path $env:ProgramData CustomScripts)
    If (!(Test-Path $InstallerFolder)) {
        New-Item -Path $InstallerFolder -ItemType Directory -Force -Confirm:$false
    }

    #Download WinGet MSIXBundle
    Write-Host "Not installed. Downloading WinGet..."
    $WinGetURL = "https://aka.ms/getwinget"
    $dc.DownloadFile($WinGetURL, "$InstallerFolder\Microsoft.DesktopAppInstaller_8wekyb3d8bbwe.msixbundle")

    If (Test-Path "$InstallerFolder\Microsoft.DesktopAppInstaller_8wekyb3d8bbwe.msixbundle" -PathType Leaf) {
        #Install WinGet MSIXBundle
        Try {
            Write-Host "Installing MSIXBundle for App Installer..."
            Add-AppxProvisionedPackage -Online -PackagePath "$InstallerFolder\Microsoft.DesktopAppInstaller_8wekyb3d8bbwe.msixbundle" -SkipLicense
            Write-Host "Installed MSIXBundle for App Installer" -ForegroundColor Green
        } Catch {
            Write-Host "Failed to install MSIXBundle for App Installer..." -ForegroundColor Red
            Exit 1
        }
    } Else {
        Write-Host "Failed to download or save MSIXBundle for App Installer..." -ForegroundColor Red
        Exit 1
    }
}

# lancement applis ( exemple)
$ResolveWingetPath = Resolve-Path "C:\Program Files\WindowsApps\Microsoft.DesktopAppInstaller_*_x64__8wekyb3d8bbwe"
    If ($ResolveWingetPath){
        $WingetPath = $ResolveWingetPath[-1].Path
        Write-Host "WinGet path is :`n$WingetPath" -ForegroundColor Green
    }

Set-Location $WingetPath
class WingetSoftware {
    [string]$Name
    [string]$PackageIdentifier
    [string]$Version
    [string]$AvailableVersion
    [string]$Source
}

# récupération de la liste de tous les logiciels installés avec winget list
# permet d'avoir la liste des installations hors winget pour remontée, et pour desinstallation
# ces infos ne sont pas utilisées pour la mise à jour et l'installation via winget.

[Console]::OutputEncoding = [System.Text.Encoding]::UTF8
$upgradeResult = .\winget.exe list | Out-String

$lines = $upgradeResult.Split([Environment]::NewLine)
# Find the line that starts with Name, it contains the header
$fl = 0
While (-not $lines[$fl].StartsWith("Nom")) {
    $fl++
}

# Line $i has the header, we can find char where we find ID and Version
$idStart = $lines[$fl].IndexOf("ID")
$versionStart = $lines[$fl].IndexOf("Version")
$availableStart = $lines[$fl].IndexOf("Disponible")
$sourceStart = $lines[$fl].IndexOf("Source")

# Now cycle in real package and split accordingly
$upgradeList = @()
For ($i = $fl + 1; $i -le $lines.Length; $i++) {
    $line = $lines[$i]

    If ($line.Length -gt ($versionStart + 1) -and -not $line.StartsWith('-')) {
        # when the name of the software is non-ascii, rune counting becomes messy.
        $nameWithSpaceEnding = $line.Substring(0, $idStart)
        $nameWithSpaceEndingByteCount = [System.Text.Encoding]::UTF8.GetByteCount($nameWithSpaceEnding)
        $nameOffset = ($nameWithSpaceEndingByteCount - $nameWithSpaceEnding.Length) / 2;
        $name = $line.Substring(0, $idStart - $nameOffset).Trim()
        # need to Trim() instead of TrimEnd() because somehow when run with | Out-String,
        # the winget output of name becomse ellipsized... Trim() would work around this.
        $id = $line.Substring($idStart - $nameOffset, $versionStart - $idStart).Trim()
        $version = $line.Substring($versionStart - $nameOffset, $availableStart - $versionStart).Trim()
        Try {
            $available = $line.Substring($availableStart - $nameOffset, $sourceStart - $availableStart).Trim()
            $source = $line.Substring($sourceStart - $nameOffset, $line.length + 1 - $sourceStart).Trim()
        } Catch {
            $available = ""
            $source = ""
        }
        $software = [WingetSoftware]::new()
        $software.Name = $name;
        $software.PackageIdentifier = $id;
        $software.Version = $version
        $software.AvailableVersion = $available;
        $software.Source = $source;

        $upgradeList += $software
    }
}

$list = $upgradeList | ConvertTo-Json
# write-Output $list


# export liste
.\winget.exe export --accept-source-agreements --output $env:temp\winget_export.json

$Uri = 'http://se4fs/wpkg/winget_out.php'

If (Test-Path $env:temp\winget_export.json -PathType Leaf) {
    $export = Get-Content -Raw -Path $env:temp\winget_export.json
    Write-Output "Export :"
    $export

    $Form = @{
        machine  = $env:ComputerName
        list     = $export
        action   = 'export'
    }
    Write-Output "Export form :"
    $Form

    Try {
        $uninstall = Invoke-RestMethod -Uri $Uri -Method Post -Body $Form
    } Catch {
        Write-Host "Failed to download 'to be uninstalled' apps..." -ForegroundColor Red
        Exit 1
    }

    # uninstall
    Write-Output "To be uninstalled :"
    $uninstall

    ForEach ( $app in $uninstall ) {
        .\winget.exe uninstall --exact --id $app.PackageIdentifier --silent --accept-source-agreements --scope machine
    }

    # install
    $Form = @{
        machine  = $env:ComputerName
        action   = 'install'
    }
    Write-Output "Install form :"
    $Form
    Invoke-WebRequest -Uri $Uri -Method Post -Body $Form -Out $env:temp\winget_import.json

    If (Test-Path $env:temp\winget_import.json -PathType Leaf) {
        $import = Get-Content -Raw -Path $env:temp\winget_import.json
        Write-Output "Import :"
        $import
        .\winget.exe import --accept-package-agreements --accept-source-agreements --ignore-unavailable --import-file $env:temp\winget_import.json
    } Else {
        Write-Host "Failed to download 'to be installed' JSON..." -ForegroundColor Red
        Exit 1
    }
} Else {
    Write-Host "Failed to export app list..." -ForegroundColor Red
    Exit 1
}
